package net.udby.test.db.lock;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class ConnectionSpec {
    private static final Map<String, String> driverMap;
    static {
        Map<String, String> map = new HashMap<String, String>();
        map.put("db2", "com.ibm.db2.jcc.DB2Driver");
        map.put("postgresql", "org.postgresql.Driver");
        map.put("derby", "org.apache.derby.jdbc.EmbeddedDriver");
        map.put("hsqldb", "org.hsqldb.jdbcDriver");
        map.put("oracle", "oracle.jdbc.driver.OracleDriver");
        driverMap = Collections.unmodifiableMap(map);
    }
    
    private static final ThreadLocal<ConnectionSpec> holder = new ThreadLocal<ConnectionSpec>();
    
    public final String schemaSpec;

    private final String schema;
    private final String url;
    private final String user;
    private final String password;
    private final String driverId;
    private final String driverClass;
    private Integer isolationLevel;
    
    private Connection con;
    private DatabaseMetaData dmd;

    public ConnectionSpec(String url, String user, String password) {
        this(url, user, password, null);
    }
    
    public ConnectionSpec(String url, String user, String password, String schema) {
        // register driver...
        int p = url.indexOf(':', 5);
        String s = url.substring(5, p);
        String driverClass = driverMap.get(s);
        if (driverClass == null) {
            throw new RuntimeException("No driver for url="+url);
        } else {
            try {
                Class.forName(driverClass);
            } catch (Exception e) {
                throw new RuntimeException("No driver for url="+url, e);
            }
        }
        this.url = url;
        this.user = user;
        this.password = password;
        if (schema== null || "".equals(schema)) {
            this.schemaSpec = "";
        } else {
            this.schemaSpec = schema + ".";
        }
        this.schema = schema==null ? "" : schema;
        this.driverId = s;
        this.driverClass = driverClass;
    }
    
    private ConnectionSpec(String url, String user, String password, String schema, Integer isolationLevel) {
        this(url, user, password, schema);
        this.isolationLevel = isolationLevel;
    }

    public static ConnectionSpec getInstance() {
        return holder.get();
    }
    
    public Connection openConnection() throws SQLException {
        if (this.con == null) {
            this.con = (Connection) DriverManager.getConnection(this.url, this.user, this.password);
            try {
                this.con.setAutoCommit(false);
                if (this.isolationLevel != null) {
                    this.con.setTransactionIsolation(this.isolationLevel);
                }
            } finally {
                holder.set(this);
            }
        }
        return this.con;
    }
    
    public Connection getConnection() {
        return this.con;
    }

    public DatabaseMetaData getDatabaseMetaData() throws SQLException {
        if (this.dmd == null) {
            this.dmd = openConnection().getMetaData();
        }
        return this.dmd;
    }
    
    public void close() throws SQLException {
        try {
            if (this.con != null) {
                this.con.close();
                this.con = null;
            }
        } finally {
            if (holder.get() == this) {
                holder.remove();
            }
        }
    }
    
    public static void cleanup() {
        ConnectionSpec ds = holder.get();
        if (ds != null) {
            holder.remove();
            try {
                try {
                    if (ds.con != null) {
                        ds.con.rollback();
                    }
                } finally {
                    ds.close();
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    public boolean supportsSelectForUpdate(DatabaseMetaData dm) throws SQLException {
        return dm.supportsSelectForUpdate();
    }
    public boolean supportsSelectForUpdate() throws SQLException {
        return supportsSelectForUpdate(getDatabaseMetaData());
    }
    public boolean supportsUpdatableResultSet(DatabaseMetaData dm) throws SQLException {
        return dm.supportsResultSetConcurrency(ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_UPDATABLE);
    }
    public boolean supportsUpdatableResultSet() throws SQLException {
        return supportsUpdatableResultSet(getDatabaseMetaData());
    }
    
    public List<String[]> driverInfo() throws SQLException {
        List<String[]> info = new ArrayList<String[]>();
        info.add(new String[]{"url", this.url});
        info.add(new String[]{"driverClass", this.driverClass});
        if (this.con != null) {
            DatabaseMetaData dm = getDatabaseMetaData();
            info.add(new String[]{"prod name", dm.getDatabaseProductName()});
            info.add(new String[]{"prod version", dm.getDatabaseProductVersion()});
            info.add(new String[]{"driver name", dm.getDriverName()});
            info.add(new String[]{"driver version", dm.getDriverVersion()});
            info.add(new String[]{"supports SELECT FOR UPDATE", Boolean.toString(supportsSelectForUpdate(dm))});
            info.add(new String[]{"supports updatable resultset", Boolean.toString(supportsUpdatableResultSet(dm))});
            info.add(new String[]{"supports transactions", Boolean.toString(dm.supportsTransactions())});
            info.add(new String[]{"supports READ_COMMITTED", Boolean.toString(dm.supportsTransactionIsolationLevel(Connection.TRANSACTION_READ_COMMITTED))});
            info.add(new String[]{"supports REPEATABLE_READ", Boolean.toString(dm.supportsTransactionIsolationLevel(Connection.TRANSACTION_REPEATABLE_READ))});
            info.add(new String[]{"isolation level", lookupField("TRANSACTION", Connection.class, con.getTransactionIsolation()) + " (" + con.getTransactionIsolation() + ")"});
        } else {
            info.add(new String[]{"connection", "closed"});
        }
        return Collections.unmodifiableList(info);
    }
    
    public ConnectionSpec clone() {
        return new ConnectionSpec(this.url, this.user, this.password, this.schema, this.isolationLevel);
    }

    private static String lookupField(String prefix, Class<?> clazz, int value) {
        try {
            Field[] fields = clazz.getDeclaredFields();
            for (int i = 0; i < fields.length; i++) {
                Field field = fields[i];
                if (Modifier.isStatic(field.getModifiers())) {
                    Object o = field.get(null);
                    if ((o instanceof Number) && ((Number)o).intValue()==value) {
                        return field.getName();
                    }
                }
            }
            return "";
        } catch (Exception e) {
            e.printStackTrace();
            return "???";
        }
    }

    /**
     * @return the isolationLevel
     */
    public int getIsolationLevel() {
        return this.isolationLevel;
    }

    /**
     * @param isolationLevel the isolationLevel to set
     */
    public void setIsolationLevel(int isolationLevel) {
        this.isolationLevel = isolationLevel;
    }
}