package net.udby.test.db.lock;

import java.lang.reflect.Constructor;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.CountDownLatch;

public class TestDriver implements Runnable {
    private final ConnectionSpec cs;
    private final ExcelReportGenerator report;
    private int meetingId;
    private final Set<Integer> passNumbers = Collections.synchronizedSet(new TreeSet<Integer>());
    
    public static final List<String> dbtypes = Collections.unmodifiableList(Arrays.asList(new String[]{"db2", "db2rr", "pgsql", "hsqlm", "hsqlmrr", "oracle", "derby", "derbyrr"}));

    private static final String[] DROP_SQL = {
        "drop table PASS",
        "drop table PASS_NEXT",
        "drop table MEETING"
    };
    private static final String[] CREATE_SQL = {
        "create table MEETING (id integer not null, info character varying(50) not null, primary key(id))",
        "create table PASS_NEXT (meeting_id integer not null, pass_number integer not null, primary key (meeting_id))",
        "alter table PASS_NEXT add constraint fk_pass_nxt_mtng foreign key (meeting_id)  references MEETING(id)",
        "create table PASS (id integer not null, meeting_id integer not null, pass_number integer not null, primary key(id))",
        "alter table PASS add constraint fk_pass_mtng foreign key (meeting_id) references MEETING(id)",
        "alter table PASS add constraint uk_pass_mtng_nmbr unique(meeting_id, pass_number)"
    };

    private static class RunInfo {
        final boolean selectForUpdate;
        final boolean updatableResultSet;
        final Class<? extends Runner> runClass;
        public RunInfo(boolean selectForUpdate, boolean updatableResultSet, Class<? extends Runner> runClass) {
            super();
            this.selectForUpdate = selectForUpdate;
            this.updatableResultSet = updatableResultSet;
            this.runClass = runClass;
        }
        public String msg() {
            if (!selectForUpdate && !updatableResultSet) {
                return "n/a";
            }
            String s = null;
            if (selectForUpdate) {
                s = "SFU";
            }
            if (updatableResultSet) {
                if (s != null) {
                    s += "+URS";
                } else {
                    s = "URS";
                }
            }
            return s;
        }
    };
    
    private static final RunInfo[] RUN_INFO = {
        new RunInfo(false, false, SimpleRunner.class),
        new RunInfo(false, true, UpdateRSRunner.class),
        new RunInfo(true, false, ForUpdateRunner.class),
        new RunInfo(true, true, ForUpdateRSRunner.class)
    };
    
    public TestDriver(final String dbtype, final ExcelReportGenerator report) throws SQLException {
        Thread.currentThread().setName("main:"+dbtype);
        this.cs = resolveDriverSpec(dbtype);
        this.cs.openConnection();
        this.report = report;
        report.createSheet(dbtype);
        report.driverInfo(this.cs.driverInfo());
    }

    public void run() {
        try {
            Logger.getInstance().log("TestDriver started, driver="+this.cs.getDatabaseMetaData().getDriverName()+"/"+this.cs.getDatabaseMetaData().getDriverVersion());
            // Drop all tables
            executeSilently(DROP_SQL);
            // Recreate..
            executeAll(CREATE_SQL);
            Logger.getInstance().log("Ready");
            for (int j = 0; j < 4; j++) {
                passNumbers.clear();
                createNewMeeting();
                if (RUN_INFO[j].selectForUpdate && !cs.supportsSelectForUpdate()) {
                    Logger.getInstance().log("Test " + (j+1) + " skipped, driver does not support SELECT FOR UPDATE");
                    report.testReportSkipped(j+1);
                    continue;
                }
                if (RUN_INFO[j].updatableResultSet && !cs.supportsUpdatableResultSet()) {
                    Logger.getInstance().log("Test " + (j+1) + " skipped, driver does not support updatable ResultSet");
                    report.testReportSkipped(j+1);
                    continue;                    
                }
                long t0 = System.nanoTime();
                Thread[] threads = createThreads(5, "test", RUN_INFO[j].runClass, this.cs);
                for (int i = 0; i < threads.length; i++) {
                    threads[i].join();
                }
                int rval = selectPassNextNumber();
                if (rval != threads.length * Runner.RUNS) {
                    FirstFailure.getInstance().failure(this, "Sanity check failed: expected="+(threads.length * Runner.RUNS)+", actual="+rval);
                }
                if (FirstFailure.failed()) {
                    report.testReportFailed(j+1, FirstFailure.getInstance().getMessage());
                    FirstFailure.clear();
                } else {
                    report.testReportSuccess(RUN_INFO[j].msg(), j+1, System.nanoTime() - t0);
                }
            }
        } catch (Exception e) {
            FirstFailure.getInstance().failure(this, e);
        } finally {
            ConnectionSpec.cleanup();
        }
    }

    private int selectPassNextNumber() throws SQLException {
        PreparedStatement ps = cs.getConnection().prepareStatement("select pass_number from pass_next where meeting_id=?");
        try {
            ps.setInt(1, this.meetingId);
            ResultSet rs = ps.executeQuery();
            try {
                if (rs.next()) {
                    return rs.getInt(1);
                } else {
                    return -1;
                }
            } finally {
                rs.close();
            }
        } finally {
            ps.close();
        }
    }

    private void createNewMeeting() throws SQLException {
        PreparedStatement ps = cs.getConnection().prepareStatement("insert into MEETING (id, info) values (?,?)");
        try {
            this.meetingId++;
            ps.setInt(1, this.meetingId);
            ps.setString(2, "Meeting:"+this.meetingId+" created "+new Timestamp(System.currentTimeMillis()).toString());
            if (1 != ps.executeUpdate()) {
                throw new SQLException("INSERT did not insert one row");
            }
        } catch (SQLException e) {
            ps.getConnection().rollback();
            throw e;
        } finally {
            ps.close();
        }
        ps = cs.openConnection().prepareStatement("insert into PASS_NEXT (meeting_id, pass_number) values (?,0)");
        try {
            ps.setInt(1, this.meetingId);
            if (1 != ps.executeUpdate()) {
                throw new SQLException("INSERT did not insert one row");
            }
        } catch (SQLException e) {
            ps.getConnection().rollback();
            throw e;
        } finally {
            ps.close();
        }
        cs.getConnection().commit();
    }

    private void executeAll(String[] sqls) throws SQLException {
        Statement stmt = cs.openConnection().createStatement();
        try {
            for (int i = 0; i < sqls.length; i++) {
                try {
                    stmt.execute(sqls[i]);
                } catch (SQLException e) {
                    stmt.getConnection().rollback();
                    throw e;
                }
            }
            stmt.getConnection().commit();
        } finally {
            stmt.close();
        }
    }

    private void executeSilently(String[] sqls) throws SQLException {
        Statement stmt = cs.openConnection().createStatement();
        try {
            for (int i = 0; i < sqls.length; i++) {
                try {
                    stmt.execute(sqls[i]);
                    stmt.getConnection().commit();
                } catch (SQLException ignored) {
                    System.out.println(ignored);
                    stmt.getConnection().rollback();
                }
            }
        } finally {
            stmt.close();
        }
    }

    private static ConnectionSpec resolveDriverSpec(String spc) {
        ConnectionSpec ds = null;
        if ("db2".equals(spc)) {
            ds = new ConnectionSpec("jdbc:db2:TST", "jesper", "****");
        } else if ("db2rr".equals(spc)) {
            ds = new ConnectionSpec("jdbc:db2:TST", "jesper", "****");
            ds.setIsolationLevel(Connection.TRANSACTION_REPEATABLE_READ);
        } else if ("pgsql".equals(spc)) {
            ds = new ConnectionSpec("jdbc:postgresql:db_test", "jesper", "****");
        } else if ("hsqlm".equals(spc)) {
            ds = new ConnectionSpec("jdbc:hsqldb:mem:test", "sa", "");
        } else if ("hsqlmrr".equals(spc)) {
            ds = new ConnectionSpec("jdbc:hsqldb:mem:test", "sa", "");
            ds.setIsolationLevel(Connection.TRANSACTION_REPEATABLE_READ);
        } else if ("hsqlf".equals(spc)) {
            ds = new ConnectionSpec("jdbc:hsqldb:file:test", "sa", "");
        } else if ("hsqlfrr".equals(spc)) {
            ds = new ConnectionSpec("jdbc:hsqldb:file:test", "sa", "");
            ds.setIsolationLevel(Connection.TRANSACTION_REPEATABLE_READ);
        } else if ("oracle".equals(spc)) {
            ds = new ConnectionSpec("jdbc:oracle:thin:@judby:1521:tst", "jesper", "****");
        } else if ("derby".equals(spc)) {
            ds = new ConnectionSpec("jdbc:derby:test;create=true", "", "");
        } else if ("derbyrr".equals(spc)) {
            ds = new ConnectionSpec("jdbc:derby:test;create=true", "", "");
            ds.setIsolationLevel(Connection.TRANSACTION_REPEATABLE_READ);
        }
        return ds;
    }
    
    private Thread[] createThreads(int cnt, String namePfx, Class<? extends Runner> executor, ConnectionSpec cs) throws Exception {
        Thread[] threads = new Thread[cnt];
        Constructor<? extends Runner> ctor = executor.getConstructor(TestDriver.class, ConnectionSpec.class, CountDownLatch.class, CountDownLatch.class);
        CountDownLatch waitStart = new CountDownLatch(cnt);
        CountDownLatch goGirl = new CountDownLatch(1);
        for (int i = 0; i < cnt; i++) {
            Thread t = new Thread(ctor.newInstance(this, cs.clone(), waitStart, goGirl));
            t.setName(namePfx+"-"+(i+1));
            t.start();
            threads[i] = t;
        }
        // Wait for all to be ready...
        waitStart.await();
        // Let them go !
        goGirl.countDown();
        return threads;
    }

    /**
     * @return the meetingId
     */
    public synchronized int getMeetingId() {
        return this.meetingId;
    }

    /**
     * @return the passNumbers
     */
    public Set<Integer> getPassNumbers() {
        return this.passNumbers;
    }
}
